"""
=============================================================================

This file is a part of qbm4eo.rst project.

https://github.com/FeralQubits/qbm4eo

=============================================================================

It has been modified as a part of the EuroHPC PL project funded at the Smart Growth
Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
POIR.04.02.00-00-D014/20-00.

=============================================================================
"""
import abc
import io
from itertools import islice
from typing import Any, Callable, Dict, Generator, Optional, Tuple, Union

import dimod
import numpy as np
from dimod.core.sampler import Sampler
from numpy.typing import NDArray
from scipy.special import expit
from torch import Tensor
from torch.utils.data import DataLoader
from tqdm import tqdm

INITIAL_COEFFICIENT_SCALE: float = 0.1


def infinite_dataloader_generator(
    data_loader: Union[
        DataLoader[Tuple[Any, Any]],
        Generator[Tuple[Any, Any], Any, Any],
    ]
) -> Generator[
    Tuple[Any, Any], None, None
]:  # https://www.linuxjournal.com/content/pythons-mypy-callables-and-generators
    """
    A generator that infinitely yields data batches from the given dataloader.

    :param data_loader:
        A PyTorch DataLoader initialized with the desired dataset.

    :return:
        Yields an index of a batch and the data from the data_loader.
    """
    while True:
        for batch_idx, (data, target) in enumerate(data_loader):
            yield batch_idx, (data, target)


def qubo_from_rbm_coefficients(
    weights: Union[Tensor, NDArray[np.float32]],
    v_bias: Union[Tensor, NDArray[np.float32]],
    h_bias: Union[Tensor, NDArray[np.float32]],
) -> dimod.BQM:
    """Create a QUBO problem representing RBM with given coefficients.

    :param weights:
        A square interaction matrix.
    :param v_bias:
        An N-element visible layer bias vector.
    :param h_bias:
        An M-element hidden layer bias vector.

    :return:
        A QUBO, represented as dimod.BQM with N+M variables, such that:
            -   variables 0,...,N-1 correspond to hidden layer
            -   variables 0,...,M-1 correspond to visible layer
            -   visible layer biases correspond to linear coefficients of first N
                variables
            -   hidden layer biases correspond to linear coefficients of second M
                variables
            -   weights correspond to interaction terms between first N and second M
                variables.

    .. note::
        This function does not allow for manipulating how RBM variables are mapped to
        the QUBO variables. This is not a problem if QUBO is to be used with an
        unstructured sampler. For sampler, the intended usage is to wrap it in the
        EmbeddingComposite.
    """
    linear: Dict[int, float] = {
        **{i: float(bias) for i, bias in enumerate(v_bias)},
        **{i: float(bias) for i, bias in enumerate(h_bias, start=len(v_bias))},
    }

    quadratic: Dict[Tuple[int, int], float] = {
        (i, j + len(v_bias)): float(weights[i, j])
        for i in range(len(v_bias))
        for j in range(len(h_bias))
    }

    return dimod.BQM(linear, quadratic, offset=0, vartype="BINARY")


class RBM:
    """
    A class implementing the Restricted Boltzmann Machine.
    """

    def __init__(
        self,
        num_visible: int,
        num_hidden: int,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        """
        A default constructor for the RMB class instances.

        :param num_visible:
            Number of neurons in the visible layer.
        :param num_hidden:
            Number of neurons in the hidden layer.
        :param rng:
            Random number generator to be used by the RBM.
        """
        self.num_visible: int = num_visible
        self.num_hidden: int = num_hidden
        self.rng: np.random.Generator = (
            rng if rng is not None else np.random.default_rng()
        )

        self.weights: NDArray[np.float32] = (
            self.rng.normal(size=(self.num_visible, self.num_hidden))
            * INITIAL_COEFFICIENT_SCALE
        )

        self.v_bias: NDArray[np.float32] = (
            self.rng.normal(size=self.num_visible) * INITIAL_COEFFICIENT_SCALE
        )

        self.h_bias: NDArray[np.float32] = (
            self.rng.normal(size=self.num_hidden) * INITIAL_COEFFICIENT_SCALE
        )

    def h_probs_given_v(self, v_batch: NDArray[np.float32]) -> NDArray[np.float32]:
        """
        Compute the probabilities of the hidden layer neurons being active given the
        visible layer neurons' states.

        :param v_batch:
            A batch of visible layer neurons' states.

        :return:
            A batch of probabilities of the hidden layer neurons being active.
        """
        return expit(self.h_bias + v_batch @ self.weights)

    def sample_h_given_v(self, v_batch: NDArray[np.float32]) -> NDArray[np.float32]:
        """
        Sample the hidden layer neurons' states given the visible layer neurons' states.

        :param v_batch:
            A batch of visible layer neurons' states.

        :return:
            A batch of hidden layer neurons' states.
        """
        probs: NDArray[np.float32] = self.h_probs_given_v(v_batch)

        return (self.rng.random(probs.shape) < probs).astype(float)

    def v_probs_given_h(self, h_batch: NDArray[np.float32]) -> NDArray[np.float32]:
        """
        Compute the probabilities of the visible layer neurons' states given the hidden
        layer neurons' states.

        :param h_batch:
            A batch of hidden layer neurons' states.

        :return:
            A batch of probabilities of the visible layer neurons' states.
        """
        return expit(self.v_bias + h_batch @ self.weights.T)

    def sample_v_given_h(self, h_batch: NDArray[np.float32]) -> NDArray[np.float32]:
        """
        Sample the visible layer neurons' states given the hidden layer neurons' states.

        :param h_batch:
            A batch of hidden layer neurons' states.

        :return:
            A batch of visible layer neurons' states.
        """
        probs: NDArray[np.float32] = self.v_probs_given_h(h_batch)
        return (self.rng.random(probs.shape) < probs).astype(float)

    def reconstruct(self, v_batch: NDArray[np.float32]) -> NDArray[np.float32]:
        """
        Reconstruct the visible layer neurons' states given the visible layer neurons'
        states.

        :param v_batch:
            A batch of visible layer neurons' states.

        :return:
            A batch of reconstructed visible layer neurons' states.
        """
        return self.v_probs_given_h(self.h_probs_given_v(v_batch))

    def save(self, file: Union[str, io.BytesIO]) -> None:
        """
        Save the RBM to a file.

        :param file:
            A file to save the RBM to.
        """
        np.savez(file, v_bias=self.v_bias, h_bias=self.h_bias, weights=self.weights)

    @classmethod
    def load(cls, file: Union[str, io.BytesIO]) -> "RBM":
        """
        Load the RBM from a file.

        :param file:
            A file to load the RBM from.

        :return:
            A loaded RBM.
        """
        data: Dict[str, NDArray[np.float32]] = np.load(file)

        weights: NDArray[np.float32] = data["weights"]
        v_bias: NDArray[np.float32] = data["v_bias"]
        h_bias: NDArray[np.float32] = data["h_bias"]

        rbm: "RBM" = cls(num_visible=len(v_bias), num_hidden=len(h_bias))
        rbm.weights = weights
        rbm.v_bias = v_bias
        rbm.h_bias = h_bias
        return rbm


class RBMTrainer:
    """
    A base class for implementing the RBM training algorithms.

    TODO TR:    There are some nasty casting between np.matrix and np.ndarray that
                is there to ensure proper transposition or multiplications when one
                of the matrix dimension is 1. This probable could be done more
                elegantly.
    """

    def __init__(self, num_steps: int) -> None:
        """
        A default constructor for the RBMTrainer class instances.

        :param num_steps:
            Number of training steps.
        """
        self.num_steps: int = num_steps

    def fit(
        self,
        rbm: RBM,
        data_loader: Union[
            DataLoader[Tuple[Any, Any]],
            Generator[Tuple[Any, Any], Any, Any],
        ],
        callback: Callable[[int, RBM, float], None] = None,
    ) -> None:
        """
        Fits the RBM to the data.

        :param rbm:
            An RBM to be trained.
        :param data_loader:
            A data loader.
        :param callback:
            A callback function to be called after each training step.
        """
        for i, (_, (batch, _)) in enumerate(
            pbar := tqdm(
                islice(infinite_dataloader_generator(data_loader), self.num_steps),
                total=self.num_steps,
            )
        ):
            batch = np.matrix(batch.detach().cpu().numpy().squeeze())
            self.training_step(rbm, batch)
            loss: float = (
                (np.array(batch - rbm.reconstruct(batch)) ** 2).sum()
                / batch.shape[0]
                / batch.shape[1]
            )

            pbar.set_postfix(loss=loss)

            if callback is not None:
                callback(i, rbm, loss)

    @abc.abstractmethod
    def training_step(self, rbm: RBM, batch: NDArray[np.float32]) -> None:
        """
        A single training step.

        :param rbm:
            An RBM to be trained.
        :param batch:
            A batch of data.
        """
        pass


class AnnealingRBMTrainer(RBMTrainer):
    """
    A class for training the RBM using the annealing samplers (D-Wave of CIMSampler).
    """

    def __init__(
        self,
        num_steps: int,
        sampler: Sampler,
        qubo_scale: float = 1.0,
        learning_rate: float = 0.01,
        **sampler_kwargs: Dict[str, Any],
    ) -> None:
        """
        A default constructor for the AnnealingRBMTrainer class instances.

        :param num_steps:
            Number of training steps.
        :param sampler:
            A sampler to be used during the training.
        :param qubo_scale:
            A scaling factor for the QUBO.
        :param learning_rate:
            A learning rate.
        :param sampler_kwargs:
            Additional keyword arguments to be passed to the sampler.
        """
        super().__init__(num_steps)
        self.sampler: Sampler = sampler
        self.sampler_kwargs: Dict[str, Any] = sampler_kwargs
        self.qubo_scale: float = qubo_scale
        self.learning_rate: float = learning_rate

    def training_step(self, rbm: RBM, batch: NDArray[np.float32]) -> None:
        """
        A single training step.

        :param rbm:
            An RBM to be trained.
        :param batch:
            The RBMs' visible layer neurons' states.
        """
        # Conditional probabilities given visible batch input
        hidden: NDArray[np.float32] = rbm.h_probs_given_v(batch)
        # Construct QUBO from this RBM
        bqm: dimod.BQM = qubo_from_rbm_coefficients(rbm.weights, rbm.v_bias, rbm.h_bias)
        # Scaling to compensate the temperature difference. Strangely, it seems
        # that in dimod this operation has to be done in place.
        bqm.scale(self.qubo_scale)
        # Take a sample of the same size as batch, extract only visible and hidden
        # variables. If the sampler supports num_reads, use it, otherwise repeat
        # the sampling for each data point in the batch.
        if "num_reads" in self.sampler.parameters:
            sample = self.sampler.sample(
                bqm, num_reads=len(batch), **self.sampler_kwargs
            ).record["sample"]
        else:
            sample = dimod.concatenate(
                [
                    self.sampler.sample(bqm, **self.sampler_kwargs)
                    for _ in range(len(batch))
                ]
            ).record["sample"]
        # Split, remembering that first variables correspond to hidden layer.
        sample_v = sample[:, : rbm.num_visible]
        sample_h = sample[:, rbm.num_visible :]
        # Update weights.
        rbm.weights += (
            self.learning_rate * (batch.T @ hidden - sample_v.T @ sample_h) / len(batch)
        )
        # And biases
        a = np.asarray(batch - sample_v).sum(axis=0)

        print(type(a))

        rbm.v_bias += (
            self.learning_rate * np.asarray(batch - sample_v).sum(axis=0).squeeze()
        )
        rbm.h_bias += (
            self.learning_rate * np.asarray(hidden - sample_h).sum(axis=0).squeeze()
        )


class CD1Trainer(RBMTrainer):
    """
    A class for training the RBM using the Contrastive Divergence 1 algorithm.
    """

    def __init__(self, num_steps: int, learning_rate: float = 0.01):
        """
        A default constructor for the CD1Trainer class instances.

        :param num_steps:
            Number of training steps.
        :param learning_rate:
            A learning rate.
        """
        super().__init__(num_steps)
        self.learning_rate = learning_rate

    def training_step(self, rbm: RBM, batch: NDArray[np.float32]) -> None:
        """
        A single training step.

        :param rbm:
            An RBM to be trained.
        :param batch:
            The RBMs' visible layer neurons' states.
        """
        # Conditional probabilities given visible batch input
        hidden_1: NDArray[np.float32] = np.matrix(rbm.h_probs_given_v(batch))

        # Propagate hidden -> visible -> hidden again
        visible_2: NDArray[np.float32] = rbm.v_probs_given_h(hidden_1)
        hidden_2: NDArray[np.float32] = np.matrix(rbm.h_probs_given_v(visible_2))

        # Update weights
        rbm.weights += np.array(
            self.learning_rate
            * (np.matrix(batch).T @ hidden_1 - np.matrix(visible_2).T @ hidden_2)
            / len(batch)
        )
        # And biases
        rbm.v_bias += self.learning_rate * np.array(batch - visible_2).sum(axis=0)
        rbm.h_bias += self.learning_rate * np.array(hidden_1 - hidden_2).sum(axis=0)



"""
=============================================================================

This file is a part of qbm4eo.rst project.

https://github.com/FeralQubits/qbm4eo

=============================================================================

It has been modified as a part of the EuroHPC PL project funded at the Smart Growth
Operational Programme 2014-2020, Measure 4.2 under the grant agreement no.
POIR.04.02.00-00-D014/20-00.

=============================================================================
"""

from typing import Any, Dict, Optional, Tuple

import torch
from torch import nn


class QuantizerFunc(torch.autograd.Function):
    """
    A class implementing the quantizer function.

    :note:
        This class is used to implement the quantizer function as a PyTorch
        autograd function. This is done to allow the quantizer function to be
        used in the forward pass of the LBAE model.

    :note:
        The forward and backward methods do not match signatures of their base methods.
    """

    # noinspection PyMethodOverriding
    @staticmethod
    def forward(ctx, f_input, dropout: int = 0):  # type: ignore
        """
        A method implementing the forward pass of the QuantizerFunc class.

        :param ctx:
            A context object.
        :param f_input:
            The input tensor.
        :param dropout:
            The dropout rate.

        :return:
            The quantized output tensor.
        """
        del dropout
        x = torch.sign(f_input)
        x[x == 0] = 1
        return x

    # noinspection PyMethodOverriding
    @staticmethod
    def backward(ctx, grad_output):  # type: ignore
        """
        A method implementing the backward pass of the QuantizerFunc class.

        :param ctx:
            A context object.
        :param grad_output:
            The gradient of the output tensor.

        :return:
            Clone of the gradient output.
        """
        grad_input = grad_output.clone()
        return grad_input, None


class ResBlockConvPart(nn.Module):
    """
    A class implementing a single part of a residual block for convolutional
    layers.
    """

    def __init__(
        self,
        channels: int,
        *args: Dict[str, Any],
        negative_slope: float = 0.02,
        bias: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the ResBlockConvPart class.

        :param channels:
            The number of channels of the input and output tensors.
        :param negative_slope:
            The negative slope of the LeakyReLU activation function.
        :param bias:
            A boolean flag indicating whether to use a bias term in the
            convolutional layer.
        :param args:
            Additional positional arguments.
        :param kwargs:
            Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.subnet: nn.Sequential = nn.Sequential(
            nn.LeakyReLU(negative_slope),
            nn.Conv2d(
                channels, channels, kernel_size=3, stride=1, padding=1, bias=bias
            ),
            nn.BatchNorm2d(channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        A method implementing the forward pass of the ResBlockConvPart class.

        :param x:
            The input tensor.
        """
        return self.subnet(x)


class ResBlockConv(nn.Module):
    """
    A class implementing a residual block for convolutional layers.
    """

    def __init__(
        self,
        channels: int,
        *args: Dict[str, Any],
        in_channels: Optional[int] = None,
        negative_slope: float = 0.02,
        bias: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the ResBlockConv class.

        :param channels:
            The number of channels of the output tensors.
        :param in_channels:
            The number of channels of the input tensor. If None, then the
            number of channels of the input tensor is assumed to be equal to
            the number of channels of the output tensor.
        :param negative_slope:
            The negative slope of the LeakyReLU activation function.
        :param bias:
            A boolean flag indicating whether to use a bias term in the
            convolutional layer.
        """

        super().__init__(*args, **kwargs)

        if in_channels is None:
            in_channels = channels

        self.initial_block: nn.Sequential = nn.Sequential(
            nn.Conv2d(
                in_channels, channels, kernel_size=4, stride=2, padding=1, bias=bias
            ),
            nn.BatchNorm2d(channels),
        )

        self.middle_block: nn.Sequential = nn.Sequential(
            ResBlockConvPart(channels, negative_slope=negative_slope, bias=bias),
            ResBlockConvPart(channels, negative_slope=negative_slope, bias=bias),
        )
        self.negative_slope: float = negative_slope

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        A method implementing the forward pass of the ResBlockConv class.

        :param x:
            The input tensor.
        """
        x = self.initial_block(x)
        y: torch.Tensor = x
        x = self.middle_block(x)
        return nn.functional.leaky_relu(x + y, self.negative_slope)


class LBAEEncoder(nn.Module):
    """
    A class implementing the encoder of the LBAE model.
    """

    def __init__(
        self,
        input_size: Tuple[int, ...],
        out_channels: int,
        latent_space_size: int,
        num_layers: int,
        quantize: bool,
        *args: Dict[str, Any],
        negative_slope: float = 0.02,
        bias: bool = False,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        A default constructor for the LBAEEncoder class.

        :param input_size:
            The size of the input image.
        :param out_channels:
            The number of channels in the output image.
        :param latent_space_size:
            The size of the latent space.
        :param num_layers:
            The number of layers in the encoder.

        """
        super().__init__(*args, **kwargs)

        layers = [
            nn.Conv2d(
                input_size[0],
                out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=bias,
            ),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(negative_slope),
        ]
        for i in range(num_layers):
            if i == 0:
                new_layer = ResBlockConv(out_channels)
            else:
                new_layer = ResBlockConv(
                    channels=2**i * out_channels,
                    in_channels=2 ** (i - 1) * out_channels,
                )
            layers.append(new_layer)

        layers.append(
            nn.Conv2d(
                2 ** (num_layers - 1) * out_channels,
                2**num_layers * out_channels,
                kernel_size=4,
                stride=2,
                bias=bias,
                padding=1,
            )
        )
        self.net = nn.Sequential(*layers)

        final_res = (
            input_size[1] // 2 ** (num_layers + 1),
            input_size[2] // 2 ** (num_layers + 1),
        )
        final_channels = 2**num_layers * out_channels
        self.final_conv_size = (final_channels, *final_res)
        lin_in_size = final_channels * final_res[0] * final_res[1]
        self.linear = nn.Linear(lin_in_size, latent_space_size)

        self.quantize = quantize
        self.quant = QuantizerFunc.apply

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        A method implementing the forward pass of the LBAEEncoder class.

        :param x:
            The input tensor.

        :return:
            The output tensor.
        """
        x = self.net(x)
        x = x.view(x.size(0), -1)
        x = self.linear(x)
        x = torch.tanh(x)

        if self.quantize:
            xq: torch.Tensor = self.quant(x)
        else:
            xq = x

        err_quant: torch.Tensor = torch.abs(x - xq)

        x = xq

        return x, err_quant.sum() / (x.size(0) * x.size(1))
